#!/usr/bin/python3
# Prunes read threads with no activity.
# Copyright 2016-2021, John "Sean" Greenslade

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <https://www.gnu.org/licenses/>. 


# Newest message must be at least this old to delete the thread.
newestmessagedays = 14

# Maximum number of threads to delete per run.
deletionlimit = 50
#deletionlimit = 2


import sys
import os
import time
import email
import shutil
import uuid
from collections import namedtuple
from dataclasses import dataclass,field


if len(sys.argv) != 3:
    print("Error: Directory must be passed.")
    exit(1)

pdir = sys.argv[1]
trashdir = sys.argv[2]

if ( os.path.isdir(os.path.join(pdir, "cur")) != True or
     os.path.isdir(os.path.join(pdir, "new")) != True ):
    print("Error: Source directory must be a maildir.")
    exit(1)

if ( os.path.isdir(os.path.join(trashdir, "cur")) != True or
     os.path.isdir(os.path.join(trashdir, "new")) != True ):
    print("Error: Trash directory must be a maildir.")
    exit(1)

currentTime = int(time.time())

Message = namedtuple("Message", ["mid", "parent", "references", "datetime", "subject", "path"])

msgCount = 0
flagged_count = 0
unread_count = 0

@dataclass
class thread_ob:
    messages: list[Message] = field(default_factory=list)
    force_keep: bool = False
    newest_ts: int = 0

    def __hash__(self): return hash(id(self))

mid_thread_lookup = {}

def merge_threads(keep_tob, remove_tob):
    if remove_tob.force_keep:
        keep_tob.force_keep = True

    keep_tob.newest_ts = max(keep_tob.newest_ts, remove_tob.newest_ts)

    for msg in remove_tob.messages:
        keep_tob.messages.append(msg)

        mid_thread_lookup[msg.mid] = keep_tob

        if msg.parent:
            mid_thread_lookup[msg.parent] = keep_tob

        for mid in msg.references.split():
            mid_thread_lookup[mid] = keep_tob


# Iterate over all messages.
print("Reading all messages...")
for mdir in [os.path.join(pdir, "cur"), os.path.join(pdir, "new")]:
    for item in os.listdir(mdir):
        if os.path.isfile(os.path.join(mdir, item)):
            msgCount += 1
            p = os.path.join(mdir, item)
            datetime = int(os.stat(p).st_mtime)

            # Get email info.
            fp = open(p, 'r', encoding = "ISO-8859-1")
            msg = email.message_from_file(fp)
            fp.close()
            messageid = msg.get("Message-ID", None)
            if not messageid or messageid == "":
                messageid = uuid.uuid4()
            parentid = msg.get("In-Reply-To", None)
            references = msg.get("References", "")
            subject = msg.get("Subject", "")

            thismsg = Message(messageid, parentid, references, datetime, subject, p)

            MIDs_to_lookup = [messageid]
            if parentid:
                MIDs_to_lookup.append(parentid)

            MIDs_to_lookup.extend(references.split())

            candidate_tob = None

            # Check every MID in this message to see what existing thread(s) it matches.
            for mid in MIDs_to_lookup:
                if mid in mid_thread_lookup:
                    if not candidate_tob:
                        candidate_tob = mid_thread_lookup[mid]
                    else: # We've seen another thread before, so compare this new thread to it.
                        new_tob = mid_thread_lookup[mid]
                        if not new_tob is candidate_tob:
                            # This is a bridge message, so new_tob and candidate_tob are actually the same thread.
                            #print("Merging.")
                            merge_threads(candidate_tob, new_tob)
                        # Else it's the same thread, so do nothing 
                # Else this is a new MID that we haven't seen before.

            if not candidate_tob:
                # We didn't find any matching threads, so make a new one.
                #print("Making new tob")
                fresh_tob = thread_ob()
                candidate_tob = fresh_tob
            # Else we found an existing thread, so just use it.

            # Add the message to the tob.
            candidate_tob.messages.append(thismsg)
            candidate_tob.newest_ts = max(candidate_tob.newest_ts, datetime)
            filename = item
            if "F" in filename.split(':')[-1]:
                flagged_count += 1
                candidate_tob.force_keep = True
            if "S" not in filename.split(':')[-1]:
                unread_count += 1
                candidate_tob.force_keep = True
            delta = currentTime - datetime
            if delta < newestmessagedays * 24 * 60 * 60:
                #print("Too new to delete")
                candidate_tob.force_keep = True

            # And add every MID to the lookup table.
            for mid in MIDs_to_lookup:
                mid_thread_lookup[mid] = candidate_tob

print("Finding unique threads...")
unique_tobs = set(mid_thread_lookup.values())
deletable_threads = []

for tob in unique_tobs:
    if not tob.force_keep:
        deletable_threads.append(tob)
    #for msg in tob.messages:

deletable_threads.sort(key = lambda deletable_threads: deletable_threads.newest_ts)

#print("Oldest", deletable_threads[0].newest_ts)
#print("Newest", deletable_threads[-1].newest_ts)
#print(deletable_threads[0])
#print(deletable_threads[-1])

message_del_count = 0
thread_del_count = 0

if len(deletable_threads) == 0:
    print("No threads meet deletion criteria.")

for tob in deletable_threads:
    if thread_del_count >= deletionlimit:
        break

    print("---")
    print("Deleting thread", tob.messages[0].subject)

    for msg in tob.messages:
        print("Moving", msg.path, "to trash...")
        if os.path.isfile(os.path.join(trashdir, "cur", os.path.basename(msg.path))):
            print("File already deleted!?")
            os.remove(msg.path)
        else:
            shutil.move(msg.path, os.path.join(trashdir, "cur"))
        message_del_count += 1
    thread_del_count += 1


print("---Summary---")
print("Total messages:", msgCount)
print("Total threads:", len(unique_tobs))
print("Deletable threads:", len(deletable_threads))
print("Flagged / unread messages:", flagged_count, "/", unread_count)
print("Total threads deleted:", thread_del_count)
print("Total messages deleted:", message_del_count)

sys.stdout.flush()
exit(0)
